import os
import unittest
from unittest import TestCase

import twined.exceptions
from octue.cloud.pub_sub.service import Service
from octue.resources.service_backends import GCPPubSubBackend


@unittest.skipUnless(
    condition=os.getenv("RUN_DEPLOYMENT_TESTS", "").lower() == "1",
    reason="'RUN_DEPLOYMENT_TESTS' environment variable is False or not present.",
)
class TestCloudRunDeployment(TestCase):
    # This is the service ID of the example service deployed to Google Cloud Run.
    EXAMPLE_SERVICE_ID = "octue.services.afbf37e3-7650-4e79-bc8e-37c0c26eae13"

    def test_cloud_run_deployment_forwards_exceptions_to_asking_service(self):
        """Test that exceptions raised in the (remote) responding service are forwarded to and raised by the asker."""
        asker = Service(backend=GCPPubSubBackend(project_name=os.environ["TEST_PROJECT_NAME"]))
        subscription, _ = asker.ask(service_id=self.EXAMPLE_SERVICE_ID, input_values={"invalid_input_data": "hello"})

        with self.assertRaises(twined.exceptions.InvalidValuesContents):
            asker.wait_for_answer(subscription)

    def test_cloud_run_deployment(self):
        """Test that the Google Cloud Run example deployment works, providing a service that can be asked questions and
        send responses.
        """
        parent = Service(backend=GCPPubSubBackend(project_name=os.environ["TEST_PROJECT_NAME"]))

        subscription, _ = parent.ask(
            service_id=self.EXAMPLE_SERVICE_ID,
            input_values={"n_iterations": 3},
        )

        answer = parent.wait_for_answer(subscription)
        self.assertEqual(answer, {"output_values": [1, 2, 3, 4, 5], "output_manifest": None})



import copy
import json
import os
import tempfile
from unittest.mock import patch

from octue import definitions, exceptions
from octue.cloud import storage
from octue.cloud.storage import GoogleCloudStorageClient
from octue.resources import Datafile, Dataset
from octue.resources.filter_containers import FilterSet
from tests import TEST_BUCKET_NAME
from tests.base import BaseTestCase
from tests.resources import create_dataset_with_two_files


class TestDataset(BaseTestCase):
    def _create_nested_cloud_dataset(self, dataset_name="a_dataset"):
        cloud_storage_client = GoogleCloudStorageClient()

        cloud_storage_client.upload_from_string(
            "[1, 2, 3]", cloud_path=storage.path.generate_gs_path(TEST_BUCKET_NAME, dataset_name, "file_0.txt")
        )

        cloud_storage_client.upload_from_string(
            "[4, 5, 6]", cloud_path=storage.path.generate_gs_path(TEST_BUCKET_NAME, dataset_name, "file_1.txt")
        )

        cloud_storage_client.upload_from_string(
            "['a', 'b', 'c']",
            cloud_path=storage.path.generate_gs_path(TEST_BUCKET_NAME, dataset_name, "sub-directory/sub_file.txt"),
        )

        cloud_storage_client.upload_from_string(
            "['blah', 'b', 'c']",
            cloud_path=storage.path.generate_gs_path(
                TEST_BUCKET_NAME, dataset_name, "sub-directory", "sub-sub-directory", "sub_sub_file.txt"
            ),
        )

    def _create_files_and_nested_subdirectories(self, directory_path):
        """Create files and nested subdirectories of files in the given directory.

        :param str directory_path: the directory to create the nested structure in
        :return list(str): the paths of the files in the directory and subdirectories
        """
        paths = [
            os.path.join(directory_path, "file_0.txt"),
            os.path.join(directory_path, "file_1.txt"),
            os.path.join(directory_path, "sub-directory", "sub_file.txt"),
            os.path.join(directory_path, "sub-directory", "sub-sub-directory", "sub_sub_file.txt"),
        ]

        os.makedirs(os.path.join(directory_path, "sub-directory", "sub-sub-directory"))

        # Create nested files in directory.
        for path, data in zip(paths, range(len(paths))):
            with open(path, "w") as f:
                f.write(str(data))

        return paths

    def test_instantiates_with_no_args(self):
        """Ensures a Datafile instantiates using only a path and generates a uuid ID"""
        Dataset()

    def test_instantiates_with_kwargs(self):
        """Ensures that keyword arguments can be used to construct the dataset initially"""
        files = [Datafile(path="path-within-dataset/a_test_file.csv")]
        resource = Dataset(files=files, labels="one two")
        self.assertEqual(len(resource.files), 1)

    def test_len(self):
        """Test that the length of a Dataset is the number of files it contains."""
        dataset = self.create_valid_dataset()
        self.assertEqual(len(dataset), len(dataset.files))

    def test_iter(self):
        """Test that iterating over a Dataset is equivalent to iterating over its files."""
        dataset = self.create_valid_dataset()
        iterated_files = {file for file in dataset}
        self.assertEqual(iterated_files, dataset.files)

    def test_add_single_file_to_empty_dataset(self):
        """Ensures that when a dataset is empty, it can be added to"""
        resource = Dataset()
        resource.add(Datafile(path="path-within-dataset/a_test_file.csv"))
        self.assertEqual(len(resource.files), 1)

    def test_add_single_file_to_existing_dataset(self):
        """Ensures that when a dataset is not empty, it can be added to"""
        files = [Datafile(path="path-within-dataset/a_test_file.csv")]
        resource = Dataset(files=files, labels="one two", tags={"a": "b"})
        resource.add(Datafile(path="path-within-dataset/a_test_file.csv"))
        self.assertEqual(len(resource.files), 2)

    def test_add_with_datafile_creation_shortcut(self):
        """Ensures that when a dataset is not empty, it can be added to"""
        resource = Dataset()
        resource.add(path="path-within-dataset/a_test_file.csv")
        self.assertEqual(len(resource.files), 1)

    def test_add_multiple_files(self):
        """Ensures that when a dataset is not empty, it can be added to"""
        files = [
            Datafile(path="path-within-dataset/a_test_file.csv"),
            Datafile(path="path-within-dataset/a_test_file.csv"),
        ]
        resource = Dataset()
        resource.add(*files)
        self.assertEqual(len(resource.files), 2)

    def test_cannot_add_non_datafiles(self):
        """Ensures that exception will be raised if adding a non-datafile object"""

        class NotADatafile:
            pass

        resource = Dataset()
        with self.assertRaises(exceptions.InvalidInputException) as e:
            resource.add(NotADatafile())

        self.assertIn("must be of class Datafile to add it to a Dataset", e.exception.args[0])

    def test_filter_catches_single_underscore_mistake(self):
        """Ensure that if the filter name contains only single underscores, an error is raised."""
        resource = Dataset(
            files=[
                Datafile(path="path-within-dataset/A_Test_file.csv"),
                Datafile(path="path-within-dataset/a_test_file.txt"),
            ]
        )

        with self.assertRaises(exceptions.InvalidInputException) as e:
            resource.files.filter(name_icontains="Test")

        self.assertIn("Invalid filter name 'name_icontains'. Filter names should be in the form", e.exception.args[0])

    def test_filter_name_contains(self):
        """Ensures that filter works with the name_contains and name_icontains lookups"""
        resource = Dataset(
            files=[
                Datafile(path="path-within-dataset/A_Test_file.csv"),
                Datafile(path="path-within-dataset/a_test_file.txt"),
            ]
        )
        files = resource.files.filter(name__icontains="Test")
        self.assertEqual(2, len(files))
        files = resource.files.filter(name__icontains="A")
        self.assertEqual(2, len(files))
        files = resource.files.filter(name__contains="Test")
        self.assertEqual(1, len(files))
        files = resource.files.filter(name__icontains="test")
        self.assertEqual(2, len(files))
        files = resource.files.filter(name__icontains="file")
        self.assertEqual(2, len(files))

    def test_filter_name_with(self):
        """Ensures that filter works with the name_endswith and name_startswith lookups"""
        resource = Dataset(
            files=[
                Datafile(path="path-within-dataset/a_my_file.csv"),
                Datafile(path="path-within-dataset/a_your_file.csv"),
            ]
        )
        files = resource.files.filter(name__starts_with="a_my")
        self.assertEqual(1, len(files))
        files = resource.files.filter(name__starts_with="a_your")
        self.assertEqual(1, len(files))
        files = resource.files.filter(name__starts_with="a_")
        self.assertEqual(2, len(files))
        files = resource.files.filter(name__starts_with="b")
        self.assertEqual(0, len(files))
        files = resource.files.filter(name__ends_with="_file.csv")
        self.assertEqual(2, len(files))
        files = resource.files.filter(name__ends_with="r_file.csv")
        self.assertEqual(1, len(files))
        files = resource.files.filter(name__ends_with="y_file.csv")
        self.assertEqual(1, len(files))
        files = resource.files.filter(name__ends_with="other.csv")
        self.assertEqual(0, len(files))

    def test_filter_by_label(self):
        """Ensures that filter works with label lookups"""
        resource = Dataset(
            files=[
                Datafile(path="path-within-dataset/a_my_file.csv", labels="one a2 b3 all"),
                Datafile(path="path-within-dataset/a_your_file.csv", labels="two a2 b3 all"),
                Datafile(path="path-within-dataset/a_your_file.csv", labels="three all"),
            ]
        )

        files = resource.files.filter(labels__contains="a")
        self.assertEqual(0, len(files))
        files = resource.files.filter(labels__contains="one")
        self.assertEqual(1, len(files))
        files = resource.files.filter(labels__contains="all")
        self.assertEqual(3, len(files))
        files = resource.files.filter(labels__any_label_starts_with="b")
        self.assertEqual(2, len(files))
        files = resource.files.filter(labels__any_label_ends_with="3")
        self.assertEqual(2, len(files))
        # files = resource.files.filter(labels__contains="hre")
        # self.assertEqual(1, len(files))

    def test_get_file_by_label(self):
        """Ensure files can be accessed by label from the dataset."""
        files = [
            Datafile(path="path-within-dataset/a_my_file.csv", labels="one a b3 all"),
            Datafile(path="path-within-dataset/a_your_file.csv", labels="two a2 b3 all"),
            Datafile(path="path-within-dataset/a_your_file.csv", labels="three all"),
        ]

        resource = Dataset(files=files)

        # Check working for single result
        self.assertIs(resource.get_file_by_label("three"), files[2])

        # Check raises for too many results
        with self.assertRaises(exceptions.UnexpectedNumberOfResultsException) as e:
            resource.get_file_by_label("all")

        self.assertIn("More than one result found", e.exception.args[0])

        # Check raises for no result
        with self.assertRaises(exceptions.UnexpectedNumberOfResultsException) as e:
            resource.get_file_by_label("billyjeanisnotmylover")

        self.assertIn("No results found for filters {'labels__contains': 'billyjeanisnotmylover'}", e.exception.args[0])

    def test_filter_name_filters_include_extension(self):
        """Ensures that filters applied to the name will catch terms in the extension"""
        files = [
            Datafile(path="path-within-dataset/a_test_file.csv"),
            Datafile(path="path-within-dataset/a_test_file.txt"),
        ]

        self.assertEqual(Dataset(files=files).files.filter(name__icontains="txt"), FilterSet({files[1]}))

    def test_filter_name_filters_exclude_path(self):
        """Ensures that filters applied to the name will not catch terms in the extension"""
        resource = Dataset(
            files=[
                Datafile(path="first-path-within-dataset/a_test_file.csv"),
                Datafile(path="second-path-within-dataset/a_test_file.txt"),
            ]
        )
        files = resource.files.filter(name__icontains="second")
        self.assertEqual(0, len(files))

    def test_hash_value(self):
        """Test hashing a dataset with multiple files gives a hash of length 8."""
        hash_ = self.create_valid_dataset().hash_value
        self.assertTrue(isinstance(hash_, str))
        self.assertTrue(len(hash_) == 8)

    def test_hashes_for_the_same_dataset_are_the_same(self):
        """Ensure the hashes for two datasets that are exactly the same are the same."""
        first_dataset = self.create_valid_dataset()
        second_dataset = copy.deepcopy(first_dataset)
        self.assertEqual(first_dataset.hash_value, second_dataset.hash_value)

    def test_serialise(self):
        """Test that a dataset can be serialised."""
        dataset = self.create_valid_dataset()
        self.assertEqual(len(dataset.to_primitive()["files"]), 2)

    def test_exists_in_cloud(self):
        """Test whether all files of a dataset are in the cloud or not can be determined."""
        self.assertFalse(self.create_valid_dataset().all_files_are_in_cloud)
        self.assertTrue(Dataset().all_files_are_in_cloud)

        files = [
            Datafile(path="gs://hello/file.txt", hypothetical=True),
            Datafile(path="gs://goodbye/file.csv", hypothetical=True),
        ]

        self.assertTrue(Dataset(files=files).all_files_are_in_cloud)

    def test_from_cloud(self):
        """Test that a Dataset in cloud storage can be accessed via (`bucket_name`, `output_directory`) and via
        `cloud_path`.
        """
        with tempfile.TemporaryDirectory() as temporary_directory:
            dataset = create_dataset_with_two_files(temporary_directory)
            dataset.tags = {"a": "b", "c": 1}

            cloud_path = storage.path.generate_gs_path(TEST_BUCKET_NAME, "a_directory", dataset.name)
            dataset.to_cloud(cloud_path)
            persisted_dataset = Dataset.from_cloud(cloud_path)

            self.assertEqual(persisted_dataset.path, f"gs://{TEST_BUCKET_NAME}/a_directory/{dataset.name}")
            self.assertEqual(persisted_dataset.id, dataset.id)
            self.assertEqual(persisted_dataset.name, dataset.name)
            self.assertEqual(persisted_dataset.hash_value, dataset.hash_value)
            self.assertEqual(persisted_dataset.tags, dataset.tags)
            self.assertEqual(persisted_dataset.labels, dataset.labels)
            self.assertEqual({file.name for file in persisted_dataset.files}, {file.name for file in dataset.files})

            for file in persisted_dataset:
                self.assertEqual(file.path, f"gs://{TEST_BUCKET_NAME}/a_directory/{dataset.name}/{file.name}")

    def test_from_cloud_with_no_datafile_metadata_file(self):
        """Test that any cloud directory can be accessed as a dataset if it has no `dataset_metadata.json` metadata
        file in it, the cloud dataset doesn't lose any information during serialization, and a metadata file is
        uploaded afterwards.
        """
        cloud_storage_client = GoogleCloudStorageClient()

        cloud_storage_client.upload_from_string(
            "[1, 2, 3]",
            cloud_path=storage.path.generate_gs_path(TEST_BUCKET_NAME, "my_dataset", "file_0.txt"),
        )

        cloud_storage_client.upload_from_string(
            "[4, 5, 6]",
            cloud_path=storage.path.generate_gs_path(TEST_BUCKET_NAME, "my_dataset", "file_1.txt"),
        )

        cloud_dataset = Dataset.from_cloud(cloud_path=f"gs://{TEST_BUCKET_NAME}/my_dataset")

        self.assertEqual(cloud_dataset.path, f"gs://{TEST_BUCKET_NAME}/my_dataset")
        self.assertEqual(cloud_dataset.name, "my_dataset")
        self.assertEqual({file.name for file in cloud_dataset.files}, {"file_0.txt", "file_1.txt"})

        for file in cloud_dataset:
            self.assertEqual(file.path, f"gs://{TEST_BUCKET_NAME}/my_dataset/{file.name}")

        # Test serialisation doesn't lose any information.
        deserialised_dataset = Dataset.deserialise(cloud_dataset.to_primitive())
        self.assertEqual(deserialised_dataset.id, cloud_dataset.id)
        self.assertEqual(deserialised_dataset.name, cloud_dataset.name)
        self.assertEqual(deserialised_dataset.path, cloud_dataset.path)
        self.assertEqual(deserialised_dataset.hash_value, cloud_dataset.hash_value)

        # Test dataset metadata file has been uploaded.
        dataset_metadata = json.loads(
            cloud_storage_client.download_as_string(
                cloud_path=storage.path.join(cloud_dataset.path, definitions.DATASET_METADATA_FILENAME)
            )
        )
        del dataset_metadata["id"]

        self.assertEqual(
            dataset_metadata,
            {
                "files": [
                    "gs://octue-test-bucket/my_dataset/file_0.txt",
                    "gs://octue-test-bucket/my_dataset/file_1.txt",
                ],
                "labels": [],
                "name": "my_dataset",
                "tags": {},
            },
        )

    def test_from_cloud_with_nested_dataset_and_no_datafile_json_file(self):
        """Test that a nested dataset is loaded from the cloud correctly."""
        self._create_nested_cloud_dataset()

        cloud_dataset = Dataset.from_cloud(cloud_path=f"gs://{TEST_BUCKET_NAME}/a_dataset", recursive=True)

        self.assertEqual(cloud_dataset.path, f"gs://{TEST_BUCKET_NAME}/a_dataset")
        self.assertEqual(cloud_dataset.name, "a_dataset")
        self.assertEqual(
            {file.name for file in cloud_dataset.files},
            {"file_0.txt", "file_1.txt", "sub_file.txt", "sub_sub_file.txt"},
        )

        # Test dataset metadata file has been uploaded.
        dataset_metadata = json.loads(
            GoogleCloudStorageClient().download_as_string(
                cloud_path=storage.path.join(cloud_dataset.path, definitions.DATASET_METADATA_FILENAME)
            )
        )
        del dataset_metadata["id"]

        self.assertEqual(
            set(dataset_metadata["files"]),
            {
                "gs://octue-test-bucket/a_dataset/file_0.txt",
                "gs://octue-test-bucket/a_dataset/file_1.txt",
                "gs://octue-test-bucket/a_dataset/sub-directory/sub_file.txt",
                "gs://octue-test-bucket/a_dataset/sub-directory/sub-sub-directory/sub_sub_file.txt",
            },
        )

    def test_to_cloud(self):
        """Test that a dataset can be uploaded to a cloud path, including all its files and the dataset's metadata."""
        with tempfile.TemporaryDirectory() as temporary_directory:
            dataset = create_dataset_with_two_files(temporary_directory)
            dataset.tags = {"a": "b", "c": 1}

            output_directory = "my_datasets"
            cloud_path = storage.path.generate_gs_path(TEST_BUCKET_NAME, output_directory, dataset.name)
            dataset.to_cloud(cloud_path)

            storage_client = GoogleCloudStorageClient()

            # Check its files have been uploaded.
            persisted_file_0 = storage_client.download_as_string(storage.path.join(cloud_path, "file_0.txt"))
            self.assertEqual(persisted_file_0, "0")

            persisted_file_1 = storage_client.download_as_string(storage.path.join(cloud_path, "file_1.txt"))
            self.assertEqual(persisted_file_1, "1")

            # Check its metadata has been uploaded.
            persisted_dataset_metadata = json.loads(
                storage_client.download_as_string(storage.path.join(cloud_path, definitions.DATASET_METADATA_FILENAME))
            )

            self.assertEqual(
                persisted_dataset_metadata["files"],
                [
                    f"gs://octue-test-bucket/my_datasets/{dataset.name}/file_0.txt",
                    f"gs://octue-test-bucket/my_datasets/{dataset.name}/file_1.txt",
                ],
            )

            self.assertEqual(persisted_dataset_metadata["tags"], dataset.tags.to_primitive())

    def test_to_cloud_with_nested_dataset_preserves_nested_structure(self):
        """Test that uploading a dataset containing datafiles in a nested directory structure to the cloud preserves
        this structure in the cloud.
        """
        with tempfile.TemporaryDirectory() as temporary_directory:
            local_paths = self._create_files_and_nested_subdirectories(temporary_directory)
            dataset = Dataset.from_local_directory(temporary_directory, recursive=True)

            upload_path = storage.path.generate_gs_path(TEST_BUCKET_NAME, "my-dataset")
            dataset.to_cloud(cloud_path=upload_path)

        uploaded_dataset = Dataset.from_cloud(cloud_path=upload_path)

        # Check that the paths relative to the dataset directory are the same in the cloud as they are locally.
        local_datafile_relative_paths = {
            path.split(temporary_directory)[-1].strip(os.path.sep).replace(os.path.sep, "/") for path in local_paths
        }

        cloud_datafile_relative_paths = {
            storage.path.split_bucket_name_from_gs_path(datafile.path)[-1].split("my-dataset/")[-1]
            for datafile in uploaded_dataset.files
        }

        self.assertEqual(cloud_datafile_relative_paths, local_datafile_relative_paths)

    def test_download_all_files(self):
        """Test that all files in a dataset can be downloaded with one command."""
        storage_client = GoogleCloudStorageClient()

        dataset_name = "another-dataset"
        storage_client.upload_from_string(
            string=json.dumps([1, 2, 3]),
            cloud_path=storage.path.generate_gs_path(TEST_BUCKET_NAME, dataset_name, "file_0.txt"),
        )
        storage_client.upload_from_string(
            string=json.dumps([4, 5, 6]),
            cloud_path=storage.path.generate_gs_path(TEST_BUCKET_NAME, dataset_name, "file_1.txt"),
        )

        dataset = Dataset.from_cloud(cloud_path=f"gs://{TEST_BUCKET_NAME}/{dataset_name}")

        with tempfile.TemporaryDirectory() as temporary_directory:
            dataset.download_all_files(local_directory=temporary_directory)

            with open(os.path.join(temporary_directory, "file_0.txt")) as f:
                self.assertEqual(f.read(), "[1, 2, 3]")

            with open(os.path.join(temporary_directory, "file_1.txt")) as f:
                self.assertEqual(f.read(), "[4, 5, 6]")

    def test_download_all_files_from_nested_dataset(self):
        """Test that all files in a nested dataset can be downloaded with one command."""
        self._create_nested_cloud_dataset("nested_dataset")

        dataset = Dataset.from_cloud(cloud_path=f"gs://{TEST_BUCKET_NAME}/nested_dataset", recursive=True)

        with tempfile.TemporaryDirectory() as temporary_directory:
            dataset.download_all_files(local_directory=temporary_directory)

            with open(os.path.join(temporary_directory, "file_0.txt")) as f:
                self.assertEqual(f.read(), "[1, 2, 3]")

            with open(os.path.join(temporary_directory, "file_1.txt")) as f:
                self.assertEqual(f.read(), "[4, 5, 6]")

            with open(os.path.join(temporary_directory, "sub-directory", "sub_file.txt")) as f:
                self.assertEqual(f.read(), "['a', 'b', 'c']")

            with open(os.path.join(temporary_directory, "sub-directory", "sub-sub-directory", "sub_sub_file.txt")) as f:
                self.assertEqual(f.read(), "['blah', 'b', 'c']")

    def test_download_all_files_from_nested_dataset_with_no_local_directory_given(self):
        """Test that, when downloading all files from a nested dataset and no local directory is given, the dataset
        structure is preserved in the temporary directory used.
        """
        self._create_nested_cloud_dataset("nested_dataset")

        dataset = Dataset.from_cloud(f"gs://{TEST_BUCKET_NAME}/nested_dataset", recursive=True)

        # Mock the temporary directory created in `Dataset.download_all_files` so we can access it for the test.
        temporary_directory = tempfile.TemporaryDirectory()

        with patch("tempfile.TemporaryDirectory", return_value=temporary_directory):
            dataset.download_all_files()

        with open(os.path.join(temporary_directory.name, "file_0.txt")) as f:
            self.assertEqual(f.read(), "[1, 2, 3]")

        with open(os.path.join(temporary_directory.name, "file_1.txt")) as f:
            self.assertEqual(f.read(), "[4, 5, 6]")

        with open(os.path.join(temporary_directory.name, "sub-directory", "sub_file.txt")) as f:
            self.assertEqual(f.read(), "['a', 'b', 'c']")

        with open(
            os.path.join(temporary_directory.name, "sub-directory", "sub-sub-directory", "sub_sub_file.txt")
        ) as f:
            self.assertEqual(f.read(), "['blah', 'b', 'c']")

    def test_from_local_directory(self):
        """Test that a dataset can be instantiated from a local nested directory ignoring its subdirectories and that
        extra keyword arguments can be provided for the dataset instantiation.
        """
        with tempfile.TemporaryDirectory() as temporary_directory:
            paths = self._create_files_and_nested_subdirectories(temporary_directory)
            dataset = Dataset.from_local_directory(temporary_directory, recursive=False, name="my-dataset")
            self.assertEqual(dataset.name, "my-dataset")

            # Check that just the top-level files from the directory are present in the dataset.
            datafile_paths = {datafile.path for datafile in dataset.files}
            self.assertEqual(datafile_paths, set(paths[:2]))

    def test_from_local_directory_recursively(self):
        """Test that a dataset can be instantiated from a local nested directory including its subdirectories."""
        with tempfile.TemporaryDirectory() as temporary_directory:
            paths = self._create_files_and_nested_subdirectories(temporary_directory)
            dataset = Dataset.from_local_directory(temporary_directory, recursive=True)

            # Check that all the files from the directory are present in the dataset.
            datafile_paths = {datafile.path for datafile in dataset.files}
            self.assertEqual(datafile_paths, set(paths))



CUSTOM_APP_RUN_MESSAGE = "This is a custom app run function"


def run(analysis, *args, **kwargs):
    """Run a mock analysis that simply prints a message.

    :param analysis:
    :return None:
    """
    print(CUSTOM_APP_RUN_MESSAGE)  # noqa:T001
    analysis.output_values = "App as a module works!"



import warnings

from octue.cloud import storage


def translate_bucket_name_and_path_in_bucket_to_cloud_path(bucket_name, path_in_bucket):
    warnings.warn(
        message=(
            "Using a bucket name and path in bucket will be deprecated soon. Please use `cloud_path` instead e.g."
            "'gs://bucket_name/path/to/file.txt'."
        ),
        category=DeprecationWarning,
    )

    return storage.path.generate_gs_path(bucket_name, path_in_bucket)



import os

from octue.resources.dataset import Dataset


def create_dataset_with_two_files(dataset_directory_path):
    """Create a dataset in the given directory containing two files.

    :param str dataset_directory_path:
    :return octue.resources.dataset.Dataset:
    """
    paths = [os.path.join(dataset_directory_path, filename) for filename in ("file_0.txt", "file_1.txt")]

    for path, data in zip(paths, range(len(paths))):
        with open(path, "w") as f:
            f.write(str(data))

    return Dataset.from_local_directory(dataset_directory_path)



import base64
import json
import logging
import os
import warnings

import google.api_core.exceptions
from google import auth
from google.cloud import storage
from google.cloud.storage.constants import _DEFAULT_TIMEOUT
from google_crc32c import Checksum

from octue.cloud.storage.path import split_bucket_name_from_gs_path
from octue.exceptions import CloudStorageBucketNotFound
from octue.migrations.cloud_storage import translate_bucket_name_and_path_in_bucket_to_cloud_path
from octue.utils.decoders import OctueJSONDecoder
from octue.utils.encoders import OctueJSONEncoder


logger = logging.getLogger(__name__)


OCTUE_MANAGED_CREDENTIALS = "octue-managed"


class GoogleCloudStorageClient:
    """A client for using Google Cloud Storage.

    :param str|google.auth.credentials.Credentials|None credentials:
    :return None:
    """

    def __init__(self, credentials=OCTUE_MANAGED_CREDENTIALS):
        warnings.simplefilter("ignore", category=ResourceWarning)

        if credentials == OCTUE_MANAGED_CREDENTIALS:
            credentials, self.project_name = auth.default()
        else:
            credentials = credentials

        self.client = storage.Client(project=self.project_name, credentials=credentials)

    def create_bucket(self, name, location=None, allow_existing=False, timeout=_DEFAULT_TIMEOUT):
        """Create a new bucket. If the bucket already exists, and `allow_existing` is `True`, do nothing; if it is
        `False`, raise an error.

        :param str name:
        :param str|None location: physical region of bucket; e.g. "europe-west6"; defaults to "US"
        :param bool allow_existing:
        :param float timeout:
        :raise google.cloud.exceptions.Conflict:
        :return None:
        """
        if allow_existing:
            if self.client.lookup_bucket(bucket_name=name, timeout=timeout) is not None:
                return

        self.client.create_bucket(bucket_or_name=name, location=location, timeout=timeout)

    def exists(self, cloud_path=None, bucket_name=None, path_in_bucket=None):
        """Check if a file exists at the given path.

        :param str|None cloud_path: full cloud path to the file (e.g. `gs://bucket_name/path/to/file.csv`)
        :return bool: `True` if the file exists
        """
        if not cloud_path:
            cloud_path = translate_bucket_name_and_path_in_bucket_to_cloud_path(bucket_name, path_in_bucket)

        blob = self._blob(cloud_path=cloud_path)
        return blob.exists()

    def upload_file(
        self,
        local_path,
        cloud_path=None,
        bucket_name=None,
        path_in_bucket=None,
        metadata=None,
        timeout=_DEFAULT_TIMEOUT,
    ):
        """Upload a local file to a Google Cloud bucket at gs://<bucket_name>/<path_in_bucket>.

        :param str local_path: path to local file
        :param str|None cloud_path: full cloud path to upload file to (e.g. `gs://bucket_name/path/to/file.csv`)
        :param dict metadata: key-value pairs to associate with the cloud file as metadata
        :param float timeout: time in seconds to allow for the upload to complete
        :return None:
        """
        if not cloud_path:
            cloud_path = translate_bucket_name_and_path_in_bucket_to_cloud_path(bucket_name, path_in_bucket)

        blob = self._blob(cloud_path)

        with open(local_path, "rb") as f:
            blob.crc32c = self._compute_crc32c_checksum(f.read())

        blob.upload_from_filename(filename=local_path, timeout=timeout)
        self._overwrite_blob_custom_metadata(blob, metadata)
        logger.debug("Uploaded %r to Google Cloud at %r.", local_path, blob.public_url)

    def upload_from_string(
        self, string, cloud_path=None, bucket_name=None, path_in_bucket=None, metadata=None, timeout=_DEFAULT_TIMEOUT
    ):
        """Upload serialised data in string form to a file in a Google Cloud bucket at
        gs://<bucket_name>/<path_in_bucket>.

        :param str string: string to upload as file
        :param str|None cloud_path: full cloud path to upload as file to (e.g. `gs://bucket_name/path/to/file.csv`)
        :param dict metadata: key-value pairs to associate with the cloud file as metadata
        :param float timeout: time in seconds to allow for the upload to complete
        :return None:
        """
        if not cloud_path:
            cloud_path = translate_bucket_name_and_path_in_bucket_to_cloud_path(bucket_name, path_in_bucket)

        blob = self._blob(cloud_path)
        blob.crc32c = self._compute_crc32c_checksum(string)
        blob.upload_from_string(data=string, timeout=timeout)
        self._overwrite_blob_custom_metadata(blob, metadata)
        logger.debug("Uploaded data to Google Cloud at %r.", blob.public_url)

    def get_metadata(self, cloud_path=None, bucket_name=None, path_in_bucket=None, timeout=_DEFAULT_TIMEOUT):
        """Get the metadata of the given file in the given bucket.

        :param str|None cloud_path: full cloud path to file (e.g. `gs://bucket_name/path/to/file.csv`)
        :param float timeout: time in seconds to allow for the request to complete
        :return dict:
        """
        if not cloud_path:
            cloud_path = translate_bucket_name_and_path_in_bucket_to_cloud_path(bucket_name, path_in_bucket)

        bucket_name, path_in_bucket = split_bucket_name_from_gs_path(cloud_path)

        bucket = self.client.get_bucket(bucket_or_name=bucket_name)
        blob = bucket.get_blob(blob_name=self._strip_leading_slash(path_in_bucket), timeout=timeout)

        if blob is None:
            return None

        custom_metadata = {}

        if blob.metadata:
            for key, value in blob.metadata.items():
                try:
                    custom_metadata[key] = json.loads(value, cls=OctueJSONDecoder)
                except json.decoder.JSONDecodeError:
                    custom_metadata[key] = value

        return {
            "custom_metadata": custom_metadata,
            "crc32c": blob.crc32c,
            "size": blob.size,
            "updated": blob.updated,
            "time_created": blob.time_created,
            "time_deleted": blob.time_deleted,
            "custom_time": blob.custom_time,
            "bucket_name": bucket_name,
            "path_in_bucket": path_in_bucket,
        }

    def overwrite_custom_metadata(self, metadata, cloud_path=None, bucket_name=None, path_in_bucket=None):
        """Overwrite the custom metadata for the given cloud file.

        :param dict metadata: key-value pairs to set as the new custom metadata
        :param str|None cloud_path: full cloud path to file (e.g. `gs://bucket_name/path/to/file.csv`)
        :return None:
        """
        if not cloud_path:
            cloud_path = translate_bucket_name_and_path_in_bucket_to_cloud_path(bucket_name, path_in_bucket)

        blob = self._blob(cloud_path)
        self._overwrite_blob_custom_metadata(blob, metadata)

    def download_to_file(
        self, local_path, cloud_path=None, bucket_name=None, path_in_bucket=None, timeout=_DEFAULT_TIMEOUT
    ):
        """Download a file to a file from a Google Cloud bucket at gs://<bucket_name>/<path_in_bucket>.

        :param str local_path: path to download to
        :param str|None cloud_path: full cloud path to download from (e.g. `gs://bucket_name/path/to/file.csv`)
        :param float timeout: time in seconds to allow for the download to complete
        :return None:
        """
        if not cloud_path:
            cloud_path = translate_bucket_name_and_path_in_bucket_to_cloud_path(bucket_name, path_in_bucket)

        blob = self._blob(cloud_path)
        self._create_intermediate_local_directories(local_path)
        blob.download_to_filename(local_path, timeout=timeout)
        logger.debug("Downloaded %r from Google Cloud to %r.", blob.public_url, local_path)

    def download_as_string(self, cloud_path=None, bucket_name=None, path_in_bucket=None, timeout=_DEFAULT_TIMEOUT):
        """Download a file to a string from a Google Cloud bucket at gs://<bucket_name>/<path_in_bucket>.

        :param str|None cloud_path: full cloud path to download from (e.g. `gs://bucket_name/path/to/file.csv`)
        :param float timeout: time in seconds to allow for the download to complete
        :return str:
        """
        if not cloud_path:
            cloud_path = translate_bucket_name_and_path_in_bucket_to_cloud_path(bucket_name, path_in_bucket)

        blob = self._blob(cloud_path)
        data = blob.download_as_bytes(timeout=timeout)
        logger.debug("Downloaded %r from Google Cloud to as string.", blob.public_url)
        return data.decode()

    def delete(self, cloud_path=None, bucket_name=None, path_in_bucket=None, timeout=_DEFAULT_TIMEOUT):
        """Delete the given file from the given bucket.

        :param str|None cloud_path: full cloud path to file to delete (e.g. `gs://bucket_name/path/to/file.csv`)
        :param float timeout: time in seconds to allow for the request to complete
        :return None:
        """
        if not cloud_path:
            cloud_path = translate_bucket_name_and_path_in_bucket_to_cloud_path(bucket_name, path_in_bucket)

        blob = self._blob(cloud_path)
        blob.delete(timeout=timeout)
        logger.debug("Deleted %r from Google Cloud.", blob.public_url)

    def scandir(
        self,
        cloud_path=None,
        bucket_name=None,
        directory_path=None,
        filter=None,
        recursive=True,
        show_directories_as_blobs=False,
        timeout=_DEFAULT_TIMEOUT,
    ):
        """Yield the blobs belonging to the given "directory" in the given bucket.

        :param str|None cloud_path: full cloud path of directory to scan (e.g. `gs://bucket_name/path/to/file.csv`)
        :param callable filter: blob filter to constrain the yielded results
        :param bool recursive: if True, include all files in the tree below the given cloud directory
        :param bool show_directories_as_blobs: if False, do not show directories as blobs (this doesn't affect inclusion of their contained files if `recursive` is True)
        :param float timeout: time in seconds to allow for the request to complete
        :yield google.cloud.storage.blob.Blob:
        """
        if not cloud_path:
            cloud_path = translate_bucket_name_and_path_in_bucket_to_cloud_path(bucket_name, directory_path)

        if filter is None:
            filter = lambda blob: True

        bucket_name, directory_path = split_bucket_name_from_gs_path(cloud_path)
        bucket = self.client.get_bucket(bucket_or_name=bucket_name)

        if not directory_path.endswith("/"):
            directory_path += "/"

        if recursive:
            blobs = bucket.list_blobs(prefix=directory_path, timeout=timeout)
        else:
            blobs = bucket.list_blobs(prefix=directory_path, delimiter="/", timeout=timeout)

        for blob in blobs:
            if show_directories_as_blobs:
                if filter(blob):
                    yield blob

            else:
                # Ensure the blob is a file (not a directory blob).
                if filter(blob) and not blob.name.endswith("/"):
                    yield blob

    def _strip_leading_slash(self, path):
        """Strip the leading slash from a path.

        :param str path:
        :return str:
        """
        return path.lstrip("/")

    def _blob(self, cloud_path=None):
        """Instantiate a blob for the given bucket at the given path. Note that this is not synced up with Google Cloud.

        :param str|None cloud_path:
        :raise octue.exceptions.CloudStorageBucketNotFound: if the bucket isn't found
        :return google.cloud.storage.blob.Blob:
        """
        bucket_name, path_in_bucket = split_bucket_name_from_gs_path(cloud_path)

        try:
            bucket = self.client.get_bucket(bucket_or_name=bucket_name)
        except google.api_core.exceptions.NotFound:
            raise CloudStorageBucketNotFound(f"The bucket {bucket_name!r} was not found.") from None

        return bucket.blob(blob_name=self._strip_leading_slash(path_in_bucket))

    def _compute_crc32c_checksum(self, string_or_bytes):
        """Compute the CRC32 checksum of the string.

        :param str|bytes string_or_bytes:
        :return str:
        """
        if isinstance(string_or_bytes, str):
            string_or_bytes = string_or_bytes.encode()

        checksum = Checksum(string_or_bytes)
        return base64.b64encode(checksum.digest()).decode("utf-8")

    def _overwrite_blob_custom_metadata(self, blob, metadata):
        """Overwrite the custom metadata for the given blob. Note that this is synced up with Google Cloud.

        :param google.cloud.storage.blob.Blob blob: Google Cloud Storage blob to update
        :param dict metadata: key-value pairs of metadata to overwrite the blob's metadata with
        :return None:
        """
        if not metadata:
            return None

        blob.metadata = self._encode_metadata(metadata)
        blob.patch()

    def _create_intermediate_local_directories(self, local_path):
        """Create intermediate directories for the given path to a local file if they don't exist.

        :param str local_path:
        :return None:
        """
        directory = os.path.dirname(os.path.abspath(local_path))
        if not os.path.exists(directory):
            os.makedirs(directory)

    def _encode_metadata(self, metadata):
        """Encode metadata as a dictionary of JSON strings.

        :param dict metadata:
        :return dict:
        """
        if not isinstance(metadata, dict):
            raise TypeError(f"Metadata for Google Cloud storage should be a dictionary; received {metadata!r}")

        return {key: json.dumps(value, cls=OctueJSONEncoder) for key, value in metadata.items()}



import logging

from cleaner import clean, read_csv_files, read_dat_file

from octue.resources import Datafile


logger = logging.getLogger(__name__)


def run(analysis, *args, **kwargs):
    """Read a time series of files from a dataset, clean them, and write a new, cleaned, dataset.

    See the "fractal" template for an introduction to the analysis object and the purpose of this 'run' function.

    Here, let's create an example application designed to clean up CSV data files produced by an instrument, in this
    case a meteorological mast.

    The aim of this example is to teach you how to use input and output file manifests in an app - so what we'll do is:
    - Use the input manifest to read a sequence of files
    - Perform a simple transformation on some of the data (as if we were doing a data cleaning process)
    - Create new files containing the cleaned data
    - Add them to the output manifest
    """
    # You can use the attached logger to record debug statements, general information, warnings or errors
    logger.info("Starting clean up of files in %s", analysis.input_manifest.absolute_path)

    # Get the configuration value for our time averaging window (or if not present, use the default specified in
    # the twine)
    time_window = (analysis.configuration_values.get("time_window", 600),)
    logger.info("Averaging window set to %ss", time_window)

    # Get the input dataset which will be read in
    input_dataset = analysis.input_manifest.get_dataset("raw_met_mast_data")

    # There are two types of files in the dataset. Metadata file(s), saved daily, and measurement files (saved hourly).
    # Because a manifest has been created, we're able to get this data out easily with the dataset filtering
    # capabilities. Let's get the metadata and the timeseries files, whilst showing off a couple of the filters.
    #
    # See the Dataset class help for more.
    metadata_file = input_dataset.get_file_by_label("meta")

    timeseries_files = input_dataset.files.filter(labels__contains="timeseries").order_by(
        "tags__sequence", check_start_value=0, check_constant_increment=1
    )
    #
    # We used these because they're special helpers - in this case ensuring that there's only one metadata file and
    # ensuring that the timeseries files come in a strictly ordered sequence.
    #
    # We could also have picked up one or more files using general filters, like so:
    #    metadata_files = input_dataset.get_files("name__icontains", filter_value="meta")
    #
    # There's generally a few ways to do it. Choose one which is likely to be most consistent - for example if your
    # filenames might be subject to change, but you have better control over the labels, rely on those.

    # At this point it's over to you, to do whatever you want with the contents of these files.
    # For this example app, we will:
    #
    #       Use a custom function to read in the strange metadata file that came with the dataset
    metadata = read_dat_file(metadata_file)
    #
    #       Read the sequence of CSV files and concatenate into a pandas dataframe (like a table)
    data = read_csv_files(timeseries_files)
    #
    #       Clean the timeseries data up
    data = clean(data, metadata["date"])

    # The twine specifies an output dataset, so it's already been created for you (although right now its empty, of
    # course, because we haven't done the processing yet)...
    output_dataset = analysis.output_manifest.get_dataset("cleaned_met_mast_data")

    # We'll add some labels, which will help to improve searchability and allow other apps, reports, users and
    # analyses to automatically find figures and use them.
    #
    # Get descriptive with labels... they are whitespace-delimited. Labels are case insensitive, and accept a-z, 0-9,
    # and hyphens which can be used literally in search and are also used to separate words in natural language search).
    # Other special characters will be stripped.
    output_dataset.labels = "met mast cleaned"

    # Create a Datafile to hold the concatenated, cleaned output data. We could put it in the current directory
    # (by leaving local_path_prefix unspecified) but it makes sense to put it in a folder specific to this output
    # dataset - doing so avoids any race conditions arising (if other instances of this application are running at the
    # same time), and avoids storage leaks, because files get cleaned up correctly.
    timeseries_datafile = Datafile(
        timestamp=None,
        path="cleaned.csv",
        path_from=output_dataset,  # Tells it where it should be stored, in this case the output dataset folder
        skip_checks=True,  # We haven't created the actual file yet, so checks would definitely fail!
        labels="timeseries",
    )

    # Write the file (now we know where to write it)
    with timeseries_datafile.open("w") as fp:
        data.to_csv(path_or_buf=fp)

    # And finally we add it to the output
    output_dataset.add(timeseries_datafile)

    # We're done! There's only one datafile in the output dataset, but you could create thousands more and append them
    # all :)
    #
    # If you're running this on your local machine, that's it - but when this code runs as an analysis in the cloud,
    # The files in the output manifest are copied into the cloud store. Their names and labels are registered in a search
    # index so your colleagues can find the dataset you've produced.



from unittest.mock import patch

from octue.essentials.monitor_messages import send_estimated_seconds_remaining, send_status_text
from octue.resources.analysis import Analysis
from tests.base import BaseTestCase


# Version 0.0.2 of the essential monitors schema (reproduced here to avoid reliance on the availability of cloud
# storage buckets).
ESSENTIAL_MONITORS_SCHEMA = {
    "title": "Essential monitoring messages",
    "description": "A generalised schema allowing a limited but useful range of monitor messages from octue services.",
    "type": "object",
    "oneOf": [
        {
            "title": "Status text",
            "description": "A short update on progress or status of the service",
            "type": "object",
            "properties": {
                "status_text": {"type": "string"},
                "date_time": {
                    "description": "The date and time of the message in ISO format",
                    "type": "string",
                    "format": "date-time",
                },
            },
            "required": ["status_text", "date_time"],
        },
        {
            "title": "Estimated seconds remaining",
            "description": "An estimate of the number of seconds yet to elapse until the question is answered, useful for long running tasks.",
            "type": "object",
            "properties": {
                "estimated_seconds_remaining": {"type": "number", "exclusiveMinimum": 0},
                "date_time": {
                    "description": "The date and time of the message in ISO format",
                    "type": "string",
                    "format": "date-time",
                },
            },
            "required": ["estimated_seconds_remaining", "date_time"],
        },
        {
            "title": "Chart",
            "description": "A plotly-compliant chart object",
            "type": "object",
            "properties": {
                "chart": {"$ref": "https://api.plot.ly/v2/plot-schema?format=json&sha1=%27%27"},
                "date_time": {
                    "description": "The date and time of the message in ISO format",
                    "type": "string",
                    "format": "date-time",
                },
            },
            "required": ["chart", "date_time"],
        },
    ],
}


class TestMonitorMessages(BaseTestCase):
    analysis = Analysis(
        twine='{"monitor_message_schema": {"$ref": "https://refs.schema.octue.com/octue/essential-monitors/0.0.2.json"}}'
    )

    def test_send_status_text(self):
        """Test that the helper can be used to send a status text monitor message that conforms to the online schema."""
        with patch("twined.twine.Twine._get_schema", return_value=ESSENTIAL_MONITORS_SCHEMA):
            send_status_text(self.analysis, "Blah", "test-service")

    def test_send_estimated_seconds_remaining(self):
        """Test that the helper can be used to send an ETA monitor message that conforms to the online schema."""
        with patch("twined.twine.Twine._get_schema", return_value=ESSENTIAL_MONITORS_SCHEMA):
            send_estimated_seconds_remaining(self.analysis, 30, "test-service")



import os
import unittest

from octue.cloud.pub_sub.service import Service
from octue.resources.service_backends import GCPPubSubBackend
from tests.base import BaseTestCase


@unittest.skipUnless(
    condition=os.getenv("RUN_DEPLOYMENT_TESTS", "").lower() == "1",
    reason="'RUN_DEPLOYMENT_TESTS' environment variable is False or not present.",
)
class TestDataflowDeployment(BaseTestCase):
    def test_sending_question_to_dataflow_streaming_job_child(self):
        """Test that a question can be sent to an existing Google Dataflow streaming job child and an answer received
        in response.
        """
        parent = Service(backend=GCPPubSubBackend(project_name=os.environ["TEST_PROJECT_NAME"]))

        subscription, _ = parent.ask(
            service_id="octue.services.c32f9dbd-7ffb-48b1-8be5-a64495a71873",
            input_values={"n_iterations": 3},
        )

        answer = parent.wait_for_answer(subscription, timeout=3600)
        self.assertEqual(answer, {"output_values": [1, 2, 3, 4, 5], "output_manifest": None})



import os

from octue.cloud import storage
from octue.exceptions import InvalidInputException


class Pathable:
    """Mixin class to enable resources to get their path location from an owner.

    For example, datasets can get their path from the Manifest they belong to.

    :param str|None path:
    :param Pathable|None path_from:
    :return None:
    """

    def __init__(self, *args, path=None, path_from=None, **kwargs):
        super().__init__(*args, **kwargs)

        if (path_from is not None) and not isinstance(path_from, Pathable):
            raise InvalidInputException(
                "paths_from argument must be an instance of an object inheriting from Pathable() mixin"
            )

        self._path_from = path_from
        self._path_is_absolute = False

        if path and path.startswith(storage.path.CLOUD_STORAGE_PROTOCOL):
            self._path_is_in_google_cloud_storage = True

        else:
            self._path_is_in_google_cloud_storage = False

        self.path = path

    @property
    def _path_prefix(self):
        """Get the path prefix (this is the absolute_path of the owner path_from object).

        Defaults to the current working directory.

        :return str:
        """
        if self._path_from is not None:
            return self._path_from.absolute_path

        if self._path_is_absolute:
            return ""

        return os.getcwd()

    @property
    def absolute_path(self):
        """Get the absolute path of this resource.

        :return str:
        """
        if self._path_is_in_google_cloud_storage:
            return storage.path.join(self._path_prefix, self._path)

        return os.path.normpath(os.path.join(self._path_prefix, self._path))

    def path_relative_to(self, path=None):
        """Get the path of this resource relative to another. If no path is provided, the current directory is used if
        the path is for a filesystem, or the bucket is used if the path is for Google Cloud.

        :return str:
        """
        if isinstance(path, Pathable):
            path = path.absolute_path

        if self._path_is_in_google_cloud_storage:
            return storage.path.relpath(self.absolute_path, start=path)

        return os.path.relpath(self.absolute_path, start=path or os.getcwd())

    @property
    def path(self):
        """Get the path of this resource.

        :return str:
        """
        return self._path

    @path.setter
    def path(self, value):
        """Set the path of this resource.

        :param Union[str, path-like] value: Path of the resource. If the resource was instantiated with a `path_from`
        object, this path must be relative. Otherwise, absolute paths are acceptable.
        :return None:
        """
        if value and value.startswith(storage.path.CLOUD_STORAGE_PROTOCOL):
            path_is_absolute = True

        else:
            value = os.path.normpath(value or ".")
            path_is_absolute = os.path.isabs(value)

        if path_is_absolute and self._path_from is not None:
            raise InvalidInputException(
                f"You cannot set an absolute path on a pathable instantiated with 'path_from'. Set a path relative to "
                f"the path_from object ({self._path_from})"
            )

        self._path_is_absolute = path_is_absolute
        self._path = value



import json
import os
import unittest.mock
from unittest import TestCase, mock

import yaml

from octue.cloud.deployment.google.answer_pub_sub_question import answer_question
from octue.exceptions import MissingServiceID
from tests.cloud.pub_sub.mocks import MockTopic
from tests.mocks import MockOpen


SERVICE_ID = "octue.services.14b124f3-8ca9-4baf-99f5-0b79179f04a6"


class TestAnswerPubSubQuestion(TestCase):
    def test_error_raised_when_no_service_id_environment_variable(self):
        """Test that a MissingServiceID error is raised if the SERVICE_ID environment variable is missing."""
        with self.assertRaises(MissingServiceID):
            answer_question(
                question={
                    "data": {},
                    "attributes": {"question_uuid": "8c859f87-b594-4297-883f-cd1c7718ef29"},
                },
                project_name="a-project-name",
            )

    def test_error_raised_when_service_id_environment_variable_is_empty(self):
        """Test that a MissingServiceID error is raised if the SERVICE_ID environment variable is empty."""
        with mock.patch.dict(os.environ, {"SERVICE_ID": ""}):
            with self.assertRaises(MissingServiceID):
                answer_question(
                    question={
                        "data": {},
                        "attributes": {"question_uuid": "8c859f87-b594-4297-883f-cd1c7718ef29"},
                    },
                    project_name="a-project-name",
                )

    def test_with_no_app_configuration_file(self):
        """Test that the `answer_question` function uses the default service and app configuration values when the
        minimal service configuration is provided with no path to an app configuration file.
        """
        with mock.patch.dict(os.environ, {"SERVICE_ID": SERVICE_ID}):
            with mock.patch(
                "octue.configuration.open",
                unittest.mock.mock_open(read_data=yaml.dump({"services": [{"name": "test-service"}]})),
            ):
                with mock.patch("octue.cloud.deployment.google.answer_pub_sub_question.Runner") as mock_runner:
                    with mock.patch("octue.cloud.pub_sub.service.Topic", new=MockTopic):
                        with mock.patch("octue.cloud.deployment.google.answer_pub_sub_question.Service"):
                            answer_question(
                                question={
                                    "data": {},
                                    "attributes": {"question_uuid": "8c859f87-b594-4297-883f-cd1c7718ef29"},
                                },
                                project_name="a-project-name",
                            )

        mock_runner.assert_called_with(
            **{
                "app_src": ".",
                "twine": "twine.json",
                "configuration_values": None,
                "configuration_manifest": None,
                "output_manifest_path": None,
                "children": None,
                "project_name": "a-project-name",
            }
        )

    def test_with_service_configuration_file_and_app_configuration_file(self):
        """Test that the `answer_question` function uses the values in the service and app configuration files if they
        are provided.
        """

        class MockOpenForConfigurationFiles(MockOpen):
            path_to_contents_mapping = {
                "octue.yaml": yaml.dump(
                    {
                        "services": [
                            {
                                "name": "test-service",
                                "app_source_path": "/path/to/app_dir",
                                "twine_path": "path/to/twine.json",
                                "app_configuration_path": "app_configuration.json",
                            }
                        ]
                    }
                ),
                "app_configuration.json": json.dumps({"configuration_values": {"hello": "configuration"}}),
            }

        with mock.patch.dict(os.environ, {"SERVICE_ID": SERVICE_ID}):
            with mock.patch("octue.configuration.open", unittest.mock.mock_open(mock=MockOpenForConfigurationFiles)):
                with mock.patch("octue.cloud.deployment.google.answer_pub_sub_question.Runner") as mock_runner:
                    with mock.patch("octue.cloud.pub_sub.service.Topic", new=MockTopic):
                        with mock.patch("octue.cloud.deployment.google.answer_pub_sub_question.Service"):
                            answer_question(
                                question={
                                    "data": {},
                                    "attributes": {"question_uuid": "8c859f87-b594-4297-883f-cd1c7718ef29"},
                                },
                                project_name="a-project-name",
                            )

        mock_runner.assert_called_with(
            **{
                "app_src": "/path/to/app_dir",
                "twine": "path/to/twine.json",
                "configuration_values": {"hello": "configuration"},
                "configuration_manifest": None,
                "output_manifest_path": None,
                "children": None,
                "project_name": "a-project-name",
            }
        )



import json
import logging
import os

import yaml


logger = logging.getLogger(__name__)


class ServiceConfiguration:
    """A class containing the details needed to configure a service.

    :param str name: the name to give the service
    :param str app_source_path: the path to the directory containing the app's source code
    :param str twine_path: the path to the twine file defining the schema for input, output, and configuration data for the service
    :param str|None app_configuration_path: the path to the app configuration file containing configuration data for the service; if this is `None`, the default application configuration is used
    :return None:
    """

    def __init__(
        self,
        name,
        app_source_path=".",
        twine_path="twine.json",
        app_configuration_path=None,
        repository_name=None,
        repository_owner=None,
        project_name=None,
        region=None,
        dockerfile_path=None,
        cloud_build_configuration_path=None,
        maximum_instances=10,
        branch_pattern="^main$",
        environment_variables=None,
        secrets=None,
        concurrency=10,
        memory="128Mi",
        cpus=1,
        minimum_instances=0,
        temporary_files_location=None,
        setup_file_path=None,
        service_account_email=None,
        machine_type=None,
        **kwargs,
    ):
        self.name = name
        self.app_source_path = app_source_path
        self.twine_path = twine_path
        self.app_configuration_path = app_configuration_path

        # Deployed services only.
        self.repository_name = repository_name
        self.repository_owner = repository_owner
        self.project_name = project_name
        self.region = region
        self.dockerfile_path = dockerfile_path
        self.provided_cloud_build_configuration_path = cloud_build_configuration_path
        self.maximum_instances = maximum_instances
        self.branch_pattern = branch_pattern
        self.environment_variables = environment_variables or []
        self.secrets = secrets or {}

        # Cloud Run services only.
        self.concurrency = concurrency
        self.memory = memory
        self.cpus = cpus
        self.minimum_instances = minimum_instances

        # Dataflow services only.
        self.temporary_files_location = temporary_files_location
        self.setup_file_path = setup_file_path
        self.service_account_email = service_account_email
        self.worker_machine_type = machine_type

        if kwargs:
            logger.warning(f"The following keyword arguments were not used by {type(self).__name__}: {kwargs!r}.")

    @classmethod
    def from_file(cls, path):
        """Load a service configuration from a file.

        :param str path:
        :return ServiceConfiguration:
        """
        with open(path) as f:
            raw_service_configuration = yaml.load(f, Loader=yaml.SafeLoader)

        logger.info("Service configuration loaded from %r.", os.path.abspath(path))

        # Ignore services other than the first for now.
        return cls(**raw_service_configuration["services"][0])


class AppConfiguration:
    """A class containing the configuration data needed to start an app as a service. The configuration data should
    conform to the service's twine schema.

    :param str|None configuration_values: values to configure the app
    :param str|None configuration_manifest: a manifest of files to configure the app
    :param str|None output_manifest_path: the path to give the output manifest
    :param str|None children: details of the children the app requires
    :return None:
    """

    def __init__(
        self,
        configuration_values=None,
        configuration_manifest=None,
        output_manifest_path=None,
        children=None,
        **kwargs,
    ):
        self.configuration_values = configuration_values
        self.configuration_manifest = configuration_manifest
        self.output_manifest_path = output_manifest_path
        self.children = children

        if kwargs:
            logger.warning(f"The following keyword arguments were not used by {type(self).__name__}: {kwargs!r}.")

    @classmethod
    def from_file(cls, path):
        """Load an app configuration from a file.

        :param str path:
        :return AppConfiguration:
        """
        with open(path) as f:
            raw_app_configuration = json.load(f)

        logger.info("App configuration loaded from %r.", os.path.abspath(path))
        return cls(**raw_app_configuration)


def load_service_and_app_configuration(service_configuration_path):
    """Load the service configuration from the given YAML file or return an empty one.

    :param str service_configuration_path: path to service configuration file
    :return dict:
    """
    service_configuration = ServiceConfiguration.from_file(service_configuration_path)
    app_configuration = AppConfiguration()

    if service_configuration.app_configuration_path:
        try:
            app_configuration = AppConfiguration.from_file(service_configuration.app_configuration_path)
        except FileNotFoundError:
            logger.info("No app configuration found.")

    return service_configuration, app_configuration



from octue.cloud.pub_sub.service import Service
from octue.resources import service_backends


BACKEND_TO_SERVICE_MAPPING = {"GCPPubSubBackend": Service}


class Child:
    """A class representing a child service that can be asked questions. It is a convenience wrapper for `Service` that
    makes the asking of questions more intuitive for Scientists and allows easier selection of backends.

    :param str name: an arbitrary name to refer to the child by (used to access it in Analysis instances and give context to log messages forwarded from the child)
    :param str id: the UUID of the child service
    :param dict backend: must include the key "name" with a value of the name of the type of backend e.g. GCPPubSubBackend and key-value pairs for any other parameters the chosen backend expects
    :return None:
    """

    def __init__(self, name, id, backend):
        self.name = name
        self.id = id

        backend_type_name = backend.pop("name")
        backend = service_backends.get_backend(backend_type_name)(**backend)
        self._service = BACKEND_TO_SERVICE_MAPPING[backend_type_name](name=f"{self.name}-parent", backend=backend)

    def ask(
        self,
        input_values=None,
        input_manifest=None,
        subscribe_to_logs=True,
        allow_local_files=False,
        handle_monitor_message=None,
        question_uuid=None,
        timeout=20,
    ):
        """Ask the child a question (i.e. send it some input value and/or a manifest and wait for it to run an analysis
        on them and return the output values). The input values given must adhere to the Twine file of the child.

        :param any input_values: the input values of the question
        :param octue.resources.manifest.Manifest|None input_manifest: the input manifest of the question
        :param bool subscribe_to_logs: if `True`, subscribe to logs from the remote service and handle them with the local log handlers
        :param bool allow_local_files: if `True`, allow the input manifest to contain references to local files - this should only be set to `True` if the serving service will have access to these local files
        :param callable|None handle_monitor_message: a function to handle monitor messages (e.g. send them to an endpoint for plotting or displaying) - this function should take a single JSON-compatible python primitive as an argument (note that this could be an array or object)
        :param float timeout: time in seconds to wait for an answer before raising a timeout error
        :param str|None question_uuid: the UUID to use for the question if a specific one is needed; a UUID is generated if not
        :raise TimeoutError: if the timeout is exceeded while waiting for an answer
        :return dict: dictionary containing the keys "output_values" and "output_manifest"
        """
        subscription, _ = self._service.ask(
            service_id=self.id,
            input_values=input_values,
            input_manifest=input_manifest,
            subscribe_to_logs=subscribe_to_logs,
            allow_local_files=allow_local_files,
            question_uuid=question_uuid,
            timeout=timeout,
        )

        return self._service.wait_for_answer(
            subscription=subscription,
            handle_monitor_message=handle_monitor_message,
            service_name=self.name,
            timeout=timeout,
        )



import json
import logging
import warnings

import twined.exceptions
from octue.cloud import storage
from octue.definitions import OUTPUT_STRANDS
from octue.exceptions import InvalidMonitorMessage
from octue.mixins import Hashable, Identifiable, Labelable, Serialisable, Taggable
from octue.resources.manifest import Manifest
from octue.utils.encoders import OctueJSONEncoder
from octue.utils.folders import get_file_name_from_strand
from twined import ALL_STRANDS, Twine


logger = logging.getLogger(__name__)


HASH_FUNCTIONS = {
    "configuration_values": Hashable.hash_non_class_object,
    "configuration_manifest": lambda manifest: manifest.hash_value,
    "input_values": Hashable.hash_non_class_object,
    "input_manifest": lambda manifest: manifest.hash_value,
}

# Map strand names to class which we expect Twined to instantiate for us
CLASS_MAP = {"configuration_manifest": Manifest, "input_manifest": Manifest, "output_manifest": Manifest}


class Analysis(Identifiable, Serialisable, Labelable, Taggable):
    """A class representing a scientific or computational analysis, holding references to all configuration, input, and
    output data.

    An Analysis instance is unique to a specific computational analysis task, however large or small, run at a specific
    time. It will be created by the task runner, which will have validated incoming data already (Analysis doesn't
    do any validation).

    It holds references to all configuration, input, and output data, logs, connections to child twins, credentials,
    etc, so should be referred to from your code to get those items.

    It's basically the "Internal API" for your data service - a single point of contact where you can get or update
    anything you need.

    Analyses are instantiated at the top level of your app/service/twin code and you can import the instantiated
    object from there (see the templates for examples)

    :param twined.Twine|str twine: Twine instance or json source
    :param callable|None handle_monitor_message: a function that sends monitor messages to the parent that requested the analysis
    :param any configuration_values: see Runner.run() for definition
    :param octue.resources.manifest.Manifest configuration_manifest: see Runner.run() for definition
    :param any input_values: see Runner.run() for definition
    :param octue.resources.manifest.Manifest input_manifest: see Runner.run() for definition
    :param dict credentials: see Runner.run() for definition
    :param dict monitors: see Runner.run() for definition
    :param any output_values: see Runner.run() for definition
    :param octue.resources.manifest.Manifest output_manifest: see Runner.run() for definition
    :param str id: Optional UUID for the analysis
    :return None:
    """

    def __init__(self, twine, handle_monitor_message=None, skip_checks=False, **kwargs):
        if isinstance(twine, Twine):
            self.twine = twine
        else:
            self.twine = Twine(source=twine)

        self._handle_monitor_message = handle_monitor_message

        strand_kwargs = {name: kwargs.pop(name, None) for name in ALL_STRANDS}

        # Values strands.
        self.input_values = strand_kwargs.get("input_values", None)
        self.configuration_values = strand_kwargs.get("configuration_values", None)
        self.output_values = strand_kwargs.get("output_values", None)

        # Manifest strands.
        self.configuration_manifest = strand_kwargs.get("configuration_manifest", None)
        self.input_manifest = strand_kwargs.get("input_manifest", None)
        self.output_manifest = strand_kwargs.get("output_manifest", None)

        # Other strands.
        self.credentials = strand_kwargs.get("credentials", None)
        self.children = strand_kwargs.get("children", None)
        self.monitors = strand_kwargs.get("monitors", None)

        self._calculate_strand_hashes(strands=strand_kwargs)
        self._skip_checks = skip_checks

        super().__init__(**kwargs)

    def send_monitor_message(self, data):
        """Send a monitor message to the parent that requested the analysis.

        :param any data: any JSON-compatible data structure
        :return None:
        """
        try:
            self.twine.validate_monitor_message(source=data)
        except twined.exceptions.InvalidValuesContents as e:
            raise InvalidMonitorMessage(e)

        if self._handle_monitor_message is None:
            logger.warning("Attempted to send a monitor message but no handler is specified.")
            return

        self._handle_monitor_message(data)

    def finalise(self, output_dir=None, save_locally=False, upload_to_cloud=False, cloud_path=None, bucket_name=None):
        """Validate and serialise the output values and manifest, optionally writing them to files and/or the manifest
        to the cloud.

        :param str output_dir: path-like pointing to directory where the outputs should be saved to file (if None, files are not written)
        :param bool save_locally:
        :param bool upload_to_cloud:
        :param str bucket_name:
        :return dict: serialised strings for values and manifest data.
        """
        serialised_strands = {}

        for output_strand in OUTPUT_STRANDS:
            logger.debug("Serialising %r", output_strand)

            attribute = getattr(self, output_strand)
            if attribute is not None:
                attribute = json.dumps(attribute, cls=OctueJSONEncoder)

            serialised_strands[output_strand] = attribute

        logger.debug("Validating serialised output json against twine")
        self.twine.validate(**serialised_strands)

        # Optionally write the serialised strands to disk.
        if save_locally:
            for output_strand in OUTPUT_STRANDS:
                if serialised_strands[output_strand] is not None:

                    filename = get_file_name_from_strand(output_strand, output_dir)
                    with open(filename, "w") as fp:
                        fp.write(serialised_strands[output_strand])
                    logger.debug("Wrote %r to file %r", output_strand, filename)

        # Optionally write the manifest to Google Cloud storage.
        if upload_to_cloud:
            if hasattr(self, "output_manifest"):
                if not cloud_path:
                    warnings.warn(
                        message=(
                            "Using a bucket name and path in bucket will be deprecated soon. Please use `cloud_path` instead e.g."
                            "'gs://bucket_name/path/to/file.txt'."
                        ),
                        category=DeprecationWarning,
                    )

                    cloud_path = storage.path.generate_gs_path(bucket_name, output_dir)

                self.output_manifest.to_cloud(cloud_path)
                logger.debug("Wrote %r to %r.", self.output_manifest, cloud_path)

        return serialised_strands

    def _calculate_strand_hashes(self, strands):
        """Calculate the hashes of the strands specified in the HASH_FUNCTIONS constant.

        :param dict strands: strand names mapped to strand data
        :return None:
        """
        for strand_name, strand_data in strands.items():
            if strand_name in HASH_FUNCTIONS:
                strand_hash_name = f"{strand_name}_hash"

                if strand_data is not None:
                    setattr(self, strand_hash_name, HASH_FUNCTIONS[strand_name](strand_data))
                else:
                    setattr(self, strand_hash_name, None)
